#hidedel:checked ~ del,
#hidedel:checked ~ * del {
    display:none;
}
#hidedel ~ #hidedel-label::before,
#hidedel ~ * #hidedel-label::before {
    content: "☐ ";
}
#hidedel:checked ~ #hidedel-label::before,
#hidedel:checked ~ * #hidedel-label::before {
    content: "☑ ";
}