@media (max-width: 767px) {
	.caniuse-status {
		opacity: .1;
		transition: opacity .1s;
	}
}
:root {
	--caniuse-bg: #EEE;
	--caniuse-shadow: #999;
	--caniuse-nosupport-text: #ccc;
	--caniuse-partialsupport-text: #666;
}
@media (prefers-color-scheme: dark) {
	:root {
		--caniuse-bg: #222;
		--caniuse-shadow: #444;
		--caniuse-nosupport-text: #666;
		--caniuse-partialsupport-text: #bbb;
	}
}
.caniuse-status {
	background: var(--caniuse-bg, #EEE);
	border-radius: .25em;
	box-shadow: 0 0 3px var(--caniuse-shadow, #999);
	color: var(--text, black);
	font: 1em sans-serif;
	hyphens: none;
	max-width: min-content;
	overflow: hidden;
	padding: .2em;
	position: absolute;
	right: 0.3em;
	top: auto;
	width: fit-content;
	word-wrap: normal;
	z-index: 8;
}
.caniuse-status.unpositioned {
	display: none;
}
.caniuse-status.overlapping-main {
	opacity: .2;
	transition: opacity .1s;
}
.caniuse-status[open] {
	opacity: 1;
	z-index: 9;
	min-width: 9em;
	width: fit-content;
}
.caniuse-status:hover {
	outline: var(--text, black) 1px solid;
	opacity: 1;
}
.caniuse-status > summary {
	font-weight: normal;
	text-align: right;
	cursor: pointer;
	display: block;
}
.caniuse-status > p {
	font-size: 0.6em;
	margin: 0;
	padding: 0;
	clear: both;
}
.caniuse-status > p + p {
	padding-top: 0.5em;
}
.caniuse-status > .support {
	display: block;
}
.caniuse-status > .support > span {
	padding: 0.2em 0;
	display: block;
	display: table;
}
.caniuse-status > .support > span.partial {
	color: var(--caniuse-partialsupport-text, #666666);
	filter: grayscale(50%);
}
.caniuse-status > .support > span.no {
	color: var(--caniuse-nosupport-text, #CCCCCC);
	filter: grayscale(100%);
}
.caniuse-status > .support > span.no::before {
	opacity: 0.5;
}
.caniuse-status > .support > span:first-of-type {
	padding-top: 0.5em;
}
.caniuse-status > .support > span > span {
	padding: 0 0.5em;
	display: table-cell;
	vertical-align: top;
}
.caniuse-status > .support > span > span:first-child {
	width: 100%;
}
.caniuse-status > .support > span > span:last-child {
	width: 100%;
	white-space: pre;
	padding: 0;
}
.caniuse-status > .support > span::before {
	content: ' ';
	display: table-cell;
	height: 1.5em;
	background: no-repeat center center;
	background-size: contain;
	text-align: right;
	font-size: 0.75em;
	font-weight: bold;
}
.caniuse-status > .support > .and_chr::before {
	background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg);
}
.caniuse-status > .support > .and_ff::before {
	background-image: url(https://resources.whatwg.org/browser-logos/firefox.png);
}
.caniuse-status > .support > .and_uc::before {
	background-image: url(https://resources.whatwg.org/browser-logos/uc.png);
}
/* UC Browser for Android */
.caniuse-status > .support > .android::before {
	background-image: url(https://resources.whatwg.org/browser-logos/android.svg);
}
.caniuse-status > .support > .bb::before {
	background-image: url(https://resources.whatwg.org/browser-logos/bb.jpg);
}
/* Blackberry Browser */
.caniuse-status > .support > .chrome::before {
	background-image: url(https://resources.whatwg.org/browser-logos/chrome.svg);
}
.caniuse-status > .support > .edge::before {
	background-image: url(https://resources.whatwg.org/browser-logos/edge.svg);
}
.caniuse-status > .support > .firefox::before {
	background-image: url(https://resources.whatwg.org/browser-logos/firefox.png);
}
.caniuse-status > .support > .ie::before {
	background-image: url(https://resources.whatwg.org/browser-logos/ie.png);
}
.caniuse-status > .support > .ie_mob::before {
	background-image: url(https://resources.whatwg.org/browser-logos/ie-mobile.svg);
}
.caniuse-status > .support > .ios_saf::before {
	background-image: url(https://resources.whatwg.org/browser-logos/safari-ios.svg);
}
.caniuse-status > .support > .op_mini::before {
	background-image: url(https://resources.whatwg.org/browser-logos/opera-mini.png);
}
.caniuse-status > .support > .op_mob::before {
	background-image: url(https://resources.whatwg.org/browser-logos/opera.svg);
}
.caniuse-status > .support > .opera::before {
	background-image: url(https://resources.whatwg.org/browser-logos/opera.svg);
}
.caniuse-status > .support > .safari::before {
	background-image: url(https://resources.whatwg.org/browser-logos/safari.png);
}
.caniuse-status > .support > .samsung::before {
	background-image: url(https://resources.whatwg.org/browser-logos/samsung.svg);
}
.caniuse-status > .caniuse {
	text-align: right;
	font-style: italic;
}
 @media (max-width: 767px) {
	.caniuse-status.wrapped {
		width: 9em;
		height: auto;
	}
	.caniuse-status:not(.wrapped) {
		height: 1em;
		max-width: min-content;
	}
	.caniuse-status.wrapped > :not(input) {
		display: block;
	}
	.caniuse-status:not(.wrapped) > :not(input) {
		display: none;
	}
}
