.domintro {
    position: relative;
    color: green;
    background: #DDFFDD;
    margin: 2.5em 0 2em 0;
    padding: 1.5em 1em 0.5em 2em;
}

.domintro dt, .domintro dt * {
    color: black;
    font-size: inherit;
}
.domintro dd {
    margin: 0.5em 0 1em 2em; padding: 0;
}
.domintro dd p {
    margin: 0.5em 0;
}
.domintro::before {
    content: 'For web developers (non-normative)';
    background: green;
    color: white;
    padding: 0.15em 0.25em;
    font-style: normal;
    position: absolute;
    top: -0.8em;
    left: -0.8em;
}
