:root {
    --ref-hint-bg: #ddd;
    --ref-hint-text: var(--text);
}
@media (prefers-color-scheme: dark) {
    :root {
        --ref-hint-bg: #222;
        --ref-hint-text: var(--text);
    }
}

.ref-hint {
    display: inline-block;
    position: absolute;
    z-index: 35;
    width: 20em;
    width: 300px;
    height: auto;
    max-height: 500px;
    overflow: auto;
    padding: 0.5em 0.5em;
    font: small Helvetica Neue, sans-serif, Droid Sans Fallback;
    background: var(--ref-hint-bg);
    color: var(--ref-hint-text);
    border: outset 0.2em;
    white-space: normal; /* in case it's moved into a pre */
}

.ref-hint * { margin: 0; padding: 0; text-indent: 0; }

.ref-hint ul { padding: 0 0 0 1em; list-style: none; }
